import sys

if sys.version_info < (3, 7):
    raise RuntimeError('Your Python version {0} is not supported, please install '
                       'Python 3.7+'.format('.'.join(map(str, sys.version_info[:3]))))

from .AsyncSberQR import AsyncSberQR
from .SberQr import SberQR
from .api import make_request, Methods
from .exceptions import (NetworkError, SberQrAPIError)

__author__ = 'bl4ckm45k'
__version__ = '2.0.1'
__email__ = 'bl4ckm45k@gmail.com'



import logging
from http import HTTPStatus

from SberQR.exceptions import SberQrAPIError, NetworkError

logger = logging.getLogger('api')


def check_result(method_name: str, content_type: str, status_code: int, body):
    """
    Checks whether `result` is a valid API response.
    A result is considered invalid if:
    - The server returned an HTTP response code other than 200
    - The content of the result is invalid JSON.
    - The method call was unsuccessful (The JSON 'ok' field equals False)

    :param method_name: The name of the method called
    :param status_code: status code
    :param content_type: content type of result
    :param body: result body
    :return: The result parsed to a JSON dictionary
    :raises ApiException: if one of the above listed cases is applicable
    """
    logger.debug('Response for %s: [%d] "%r"', method_name, status_code, body)

    if content_type != 'application/json':
        raise NetworkError(f"Invalid response with content type {content_type}: \"{body}\"")

    if HTTPStatus.OK <= status_code <= HTTPStatus.IM_USED:
        return body
    elif status_code == HTTPStatus.BAD_REQUEST:
        raise SberQrAPIError(f"{body} [{status_code}]")
    elif status_code == HTTPStatus.NOT_FOUND:
        raise SberQrAPIError(f"{body} [{status_code}]")
    elif status_code == HTTPStatus.CONFLICT:
        raise SberQrAPIError(f"{body} [{status_code}]")
    elif status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
        raise SberQrAPIError(f"{body} [{status_code}]")
    elif status_code >= HTTPStatus.INTERNAL_SERVER_ERROR:
        raise SberQrAPIError(f"{body} [{status_code}]")


async def make_request(session, method, headers, data, **kwargs):
    url = f'https://mc.api.sberbank.ru/prod/{method}'

    if method != Methods.oauth:
        async with session.post(url, json=data, headers=headers) as response:
            try:
                body = await response.json()
            except:
                body = response.text
            return check_result(method, response.content_type, response.status, body)
    else:
        async with session.post(url, data=data, headers=headers) as response:
            try:
                body = await response.json()
            except:
                body = response.text
            return check_result(method, response.content_type, response.status, body)


class Methods:
    oauth = 'tokens/v2/oauth'
    creation = 'qr/order/v3/creation'
    status = 'qr/order/v3/status'
    revocation = 'qr/order/v3/revocation'
    cancel = 'qr/order/v3/cancel'
    registry = 'qr/order/v3/registry'



from .api import check_result, Methods


def make_request(session, method, headers, data, **kwargs):
    url = f'https://mc.api.sberbank.ru/prod/{method}'

    if method != Methods.oauth:
        with session.post(url, json=data, headers=headers) as response:
            try:
                body = response.json()
            except Exception:
                body = response.text
            return check_result(method, response.headers.get('Content-Type'), response.status_code, body)
    else:
        with session.post(url, data=data, headers=headers) as response:
            try:
                body = response.json()
            except Exception:
                body = response.text
            return check_result(method, response.headers.get('Content-Type'), response.status_code, body)



import asyncio
import base64
from datetime import datetime
from logging import getLogger
from random import choices
from string import hexdigits
from typing import Optional, Union, List, Dict

import requests
from redis.client import Redis
from requests.adapters import HTTPAdapter
from urllib3.util.ssl_ import create_urllib3_context

from .api_sync import make_request, Methods
from .payload import generate_payload
from .scope import Scope
from .types import RegistryType, CancelType

logger = getLogger(__name__)


class SberQR:

    def __init__(self, member_id: str, id_qr: str, tid: str,
                 client_id: str, client_secret: str,
                 crt_file_path: str, key_file_path: str,
                 pkcs12_password: str,
                 russian_crt: str,
                 redis: Union[str, Redis] = None,
                 loop: Optional[Union[asyncio.BaseEventLoop, asyncio.AbstractEventLoop]] = None,
                 timeout: Optional[Union[int, float, requests.Timeout]] = None):
        """

        :param member_id:
        :param id_qr:
        :param tid:
        :param client_id:
        :param client_secret: l
        i]
        :param crt_file_path:
        :param key_file_path:
        :param pkcs12_password:
        :param redis:
        :param loop:
        :param timeout:
        """

        self._main_loop = loop

        self._member_id = member_id
        self._sbp_member_id = "100000000111"
        self._id_qr = id_qr
        self._tid = tid
        self._client_id = client_id
        self._client_secret = client_secret

        self._currency = "643"
        self._redis = None

        class SSLAdapter(HTTPAdapter):
            def init_poolmanager(self, *args, **kwargs):
                context = create_urllib3_context()
                context.load_cert_chain(certfile=crt_file_path, keyfile=key_file_path, password=pkcs12_password)
                context.load_verify_locations(cafile=russian_crt)
                kwargs['ssl_context'] = context
                return super().init_poolmanager(*args, **kwargs)

        self._session: Optional[requests.Session] = None
        self._https_class = SSLAdapter()
        if isinstance(redis, Redis):
            self._redis = redis
        elif isinstance(redis, str):
            self._redis = Redis(redis, decode_responses=True)

        self.timeout = timeout

    def get_new_session(self) -> requests.Session:
        session = requests.Session()
        session.mount("https://", self._https_class)
        return session

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._main_loop

    def get_session(self) -> Optional[requests.Session]:
        if self._session is None:
            self._session = self.get_new_session()

        return self._session

    def close(self):
        """
        Close all client sessions
        """
        if self._session:
            self._session.close()

    def request(self, method, headers, data):
        headers = {**headers, **{'Accept': 'application/json', 'x-ibm-client-id': self._client_id}}
        return make_request(self.get_session(), method, headers, data)

    def get_token_from_redis(self, scope):
        """
        Возвращает токен, если он не истек
        :param scope Область токена
        :return: token string
        """
        return self._redis.get(f'{self._client_id}token_{scope.value}')

    def token(self, scope: Scope):
        redis_token = self.get_token_from_redis(scope) if self._redis else None
        if redis_token is not None:
            return redis_token
        else:
            auth = base64.b64encode(f'{self._client_id}:{self._client_secret}'.encode('utf-8')).decode('utf-8')
            headers = {'Authorization': f'Basic {auth}',
                       'Content-Type': 'application/x-www-form-urlencoded',
                       'rquid': ''.join(choices(hexdigits, k=32))}
            data = {'grant_type': 'client_credentials', 'scope': scope.value}
            token_data = self.request(Methods.oauth, headers, data)
            if self._redis:
                self._redis.set(f'{self._client_id}token_{scope.value}', token_data['access_token'],
                                int(token_data['expires_in']) - 10)
            return token_data['access_token']

    def creation(self, description: str, order_sum: int, order_number: str, positions: Union[List, Dict]):
        """
        Создание заказа
        """
        dt = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {self.token(Scope.create)}', 'RqUID': rq_uid}

        rq_tm, order_create_date = dt, dt
        member_id, id_qr, currency = self._member_id, self._id_qr, self._currency

        sbp_member_id = self._sbp_member_id if self._tid == self._id_qr else None

        if isinstance(positions, dict):
            order_params_type = [positions]
        else:
            order_params_type = positions
        del positions
        payload = generate_payload(exclude=['dt', 'headers'], **locals())
        return self.request(Methods.creation, headers, payload)

    def status(self, order_id: str, partner_order_number: str):
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {self.token(Scope.status)}', 'RqUID': rq_uid}
        tid = self._tid
        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        payload = generate_payload(exclude=['headers'], **locals())
        return self.request(Methods.status, headers, payload)

    def revoke(self, order_id: str):
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {self.token(Scope.revoke)}', 'RqUID': rq_uid}

        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        payload = generate_payload(exclude=['headers'], **locals())
        return self.request(Methods.revocation, headers, payload)

    def cancel(
            self, order_id: str, operation_id: str, cancel_operation_sum: int, auth_code: str,
            operation_type: CancelType = CancelType.REVERSE, sbp_payer_id: str = None
    ):
        """
        Отмена/возврат
        """
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {self.token(Scope.cancel)}', 'RqUID': rq_uid}

        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        id_qr, tid, operation_currency = self._id_qr, self._tid, self._currency
        operation_type = operation_type.value
        payload = generate_payload(exclude=['headers'], **locals())
        return self.request(Methods.cancel, headers, payload)

    def registry(self, start_period: datetime, end_period: datetime,
                 registry_type: RegistryType = RegistryType.REGISTRY):
        """
        Запрос реестра операций
        """
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {self.token(Scope.registry)}', 'RqUID': rq_uid}
        payload = {"rqUid": rq_uid,
                   "rqTm": f'{datetime.utcnow().isoformat(timespec="seconds")}Z',
                   "idQR": self._id_qr,
                   "startPeriod": f'{start_period.isoformat(timespec="seconds")}Z',
                   "endPeriod": f'{end_period.isoformat(timespec="seconds")}Z',
                   "registryType": registry_type.value}

        return self.request(Methods.registry, headers, payload)



import logging

logger = logging.getLogger('payload')
DEFAULT_FILTER = ['self', 'cls']


def generate_payload(exclude=None, **kwargs):
    if exclude is None:
        exclude = ['headers', 'dt']
    data = dict()
    for k, v in kwargs.items():
        if k not in exclude + DEFAULT_FILTER and v is not None:
            data[k] = v
    logger.debug(data)
    return data



import asyncio
import base64
import ssl
from datetime import datetime
from logging import getLogger
from random import choices
from string import hexdigits
from typing import Optional, Type, Union, List, Dict

import aiohttp
import certifi
import ujson as json
from redis.asyncio import Redis

from .api import make_request, Methods
from .payload import generate_payload
from .scope import Scope
from .types import RegistryType, CancelType

logger = getLogger(__name__)


class AsyncSberQR:

    def __init__(self, member_id: str, id_qr: str, tid: str,
                 client_id: str, client_secret: str,
                 crt_file_path: str, key_file_path: str,
                 pkcs12_password: str,
                 russian_crt: str,
                 redis: Union[str, Redis] = None,
                 loop: Optional[Union[asyncio.BaseEventLoop, asyncio.AbstractEventLoop]] = None,
                 connections_limit: int = None,
                 timeout: Optional[Union[int, float, aiohttp.ClientTimeout]] = None):
        """

        :param member_id:
        :param id_qr:
        :param tid:
        :param client_id:
        :param client_secret: l
        i]
        :param crt_file_path:
        :param key_file_path:
        :param pkcs12_password:
        :param redis:
        :param loop:
        :param connections_limit:
        :param timeout:
        """

        self._main_loop = loop

        self._member_id = member_id
        self._sbp_member_id = "100000000111"
        self._id_qr = id_qr
        self._tid = tid
        self._client_id = client_id
        self._client_secret = client_secret
        self._redis = None

        self._currency = "643"
        ssl_context = ssl.create_default_context(cafile=certifi.where())

        ssl_context.load_cert_chain(certfile=crt_file_path,
                                    keyfile=key_file_path,
                                    password=pkcs12_password)
        ssl_context.load_verify_locations(cafile=russian_crt)

        self._session: Optional[aiohttp.ClientSession] = None
        self._connector_class: Type[aiohttp.TCPConnector] = aiohttp.TCPConnector
        self._connector_init = dict(limit=connections_limit, ssl=ssl_context)
        if isinstance(redis, Redis):
            self._redis = redis
        elif isinstance(redis, str):
            self._redis = Redis(host=redis, decode_responses=True)

        self.timeout = timeout

    async def get_new_session(self) -> aiohttp.ClientSession:
        return aiohttp.ClientSession(
            connector=self._connector_class(**self._connector_init),
            json_serialize=json.dumps
        )

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._main_loop

    async def get_session(self) -> Optional[aiohttp.ClientSession]:
        if self._session is None or self._session.closed:
            self._session = await self.get_new_session()

        if not self._session._loop.is_running():
            await self._session.close()
            self._session = await self.get_new_session()

        return self._session

    async def close(self):
        """
        Close all client sessions
        """
        if self._session:
            await self._session.close()

    async def request(self, method, headers, data):
        headers = {**headers, **{'Accept': 'application/json', 'x-ibm-client-id': self._client_id}}
        return await make_request(await self.get_session(), method, headers, data)

    async def get_token_from_redis(self, scope):
        """
        Возвращает токен, если он не истек
        :param scope Область токена
        :return: token string
        """
        return await self._redis.get(f'{self._client_id}token_{scope.value}')

    async def token(self, scope: Scope):
        redis_token = await self.get_token_from_redis(scope) if self._redis else None
        if redis_token:
            return redis_token
        else:
            auth = base64.b64encode(f'{self._client_id}:{self._client_secret}'.encode('utf-8')).decode('utf-8')
            headers = {'Authorization': f'Basic {auth}',
                       'Content-Type': 'application/x-www-form-urlencoded',
                       'rquid': ''.join(choices(hexdigits, k=32))}
            data = {'grant_type': 'client_credentials', 'scope': scope.value}
            token_data = await self.request(Methods.oauth, headers, data)
            if self._redis:
                await self._redis.set(f'{self._client_id}token_{scope.value}', token_data['access_token'],
                                      int(token_data['expires_in']) - 10)
            return token_data['access_token']

    async def creation(self, description: str, order_sum: int, order_number: str, positions: Union[List, Dict]):
        """
        Создание заказа
        """
        dt = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {await self.token(Scope.create)}', 'RqUID': rq_uid}

        rq_tm, order_create_date = dt, dt
        member_id, id_qr, currency = self._member_id, self._id_qr, self._currency

        sbp_member_id = self._sbp_member_id if self._tid == self._id_qr else None

        if isinstance(positions, dict):
            order_params_type = [positions]
        else:
            order_params_type = positions
        del positions
        payload = generate_payload(exclude=['dt', 'headers'], **locals())
        return await self.request(Methods.creation, headers, payload)

    async def status(self, order_id: str, partner_order_number: str):
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {await self.token(Scope.status)}', 'RqUID': rq_uid}
        tid = self._tid
        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        payload = generate_payload(exclude=['headers'], **locals())
        return await self.request(Methods.status, headers, payload)

    async def revoke(self, order_id: str):
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {await self.token(Scope.revoke)}', 'RqUID': rq_uid}

        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        payload = generate_payload(exclude=['headers'], **locals())
        return await self.request(Methods.revocation, headers, payload)

    async def cancel(
            self, order_id: str, operation_id: str, cancel_operation_sum: int, auth_code: str,
            operation_type: CancelType = CancelType.REVERSE, sbp_payer_id: str = None
    ):
        """
        Отмена/возврат
        """
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {await self.token(Scope.cancel)}', 'RqUID': rq_uid}

        rq_tm = f'{datetime.utcnow().isoformat(timespec="seconds")}Z'
        id_qr, tid, operation_currency = self._id_qr, self._tid, self._currency
        operation_type = operation_type.value
        payload = generate_payload(exclude=['headers'], **locals())
        return await self.request(Methods.cancel, headers, payload)

    async def registry(self, start_period: datetime, end_period: datetime,
                       registry_type: RegistryType = RegistryType.REGISTRY):
        """
        Запрос реестра операций
        """
        rq_uid = ''.join(choices(hexdigits, k=32))
        headers = {'Authorization': f'Bearer {await self.token(Scope.registry)}', 'RqUID': rq_uid}
        payload = {"rqUid": rq_uid,
                   "rqTm": f'{datetime.utcnow().isoformat(timespec="seconds")}Z',
                   "idQR": self._id_qr,
                   "startPeriod": f'{start_period.isoformat(timespec="seconds")}Z',
                   "endPeriod": f'{end_period.isoformat(timespec="seconds")}Z',
                   "registryType": registry_type.value}

        return await self.request(Methods.registry, headers, payload)


import sys

from setuptools import setup

if sys.version_info < (3, 7):
    raise RuntimeError('Your Python version {0} is not supported, please install '
                       'Python 3.7+'.format('.'.join(map(str, sys.version_info[:3]))))
requirements = ["aiohttp>=3.8.4", "certifi>=2023.11.17", "ujson>=5.9.0", "pytz==2022.1", "qrcode[pil]>=7.3.1",
                "redis>=4.2.0rc1", "urllib3~=2.0.3"]

setup(
    name='SberQR',
    version='2.0.1',
    author='bl4ckm45k',
    author_email='nonpowa@gmail.com',
    description='Библиотека для работы с SberPay QR/Плати QR.',
    long_description_content_type="text/markdown",
    url="https://github.com/bl4ckm45k/SberQR",
    license="MIT",
    packages=['SberQR'],
    install_requires=requirements,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Office/Business :: Financial'

    ],
    python_requires='>=3.7'
)



from enum import Enum


class Scope(Enum):
    create = 'https://api.sberbank.ru/qr/order.create'
    status = 'https://api.sberbank.ru/qr/order.status'
    revoke = 'https://api.sberbank.ru/qr/order.revoke'
    cancel = 'https://api.sberbank.ru/qr/order.cancel'
    registry = 'auth://qr/order.registry'
    notify = 'auth://qr/order.notify'
